import { ARIARoleDefinitionKey } from 'aria-query';
import { IsInaccessibleOptions } from 'dom-accessibility-api';

declare class StaticText {
    text: string;
    constructor(text: string);
    toString(): string;
}

type AsNonLandmarkRoles = 'HeaderAsNonLandmark' | 'FooterAsNonLandmark';
type VirtualRoles = 'Abbr' | 'Audio' | 'Canvas' | 'Details' | 'DescriptionListDetails' | 'DescriptionList' | 'DescriptionListTerm' | 'DisclosureTriangle' | 'EmbeddedObject' | 'Figcaption' | 'PluginObject' | 'LabelText' | 'LineBreak' | 'Video';
type ARIARoleDefinitionKeyExtended = ARIARoleDefinitionKey | AsNonLandmarkRoles | VirtualRoles;
type TextMatcherFunction = (content: string, element: Element | null) => boolean;
type TextMatcher = string | number | RegExp | TextMatcherFunction;
type A11yTreeNodeContext = {
    isListSubtree?: boolean;
    isClosedDetailsSubtree?: boolean;
    isNonLandmarkSubtree?: boolean;
    isInaccessibleOptions?: IsInaccessibleOptions;
};
type A11yTreeNodeState = {
    busy: boolean;
    checked?: boolean;
    current: string | boolean;
    disabled: boolean;
    expanded?: boolean;
    pressed?: boolean;
    selected?: boolean;
};
type A11yTreeNodeQueries = {
    level?: number;
    value: {
        min?: number;
        max?: number;
        now?: number;
        text?: string;
    };
};
type A11yTreeNode = {
    element: HTMLElement;
    role: HTMLElement['role'] | undefined;
    name: string;
    description: string;
    state: A11yTreeNodeState;
    queries: A11yTreeNodeQueries;
    children?: (A11yTreeNode | StaticText)[];
};
type A11yTreeNodeStateMatch = Partial<A11yTreeNodeState>;
type A11yTreeNodeQueriesMatch = {
    level?: number;
    value?: {
        min?: number;
        max?: number;
        now?: number;
        text?: TextMatcher;
    };
};
type A11yTreeNodeMatch = {
    role?: HTMLElement['role'];
    name?: TextMatcher;
    description?: TextMatcher;
    state?: A11yTreeNodeStateMatch;
    queries?: A11yTreeNodeQueriesMatch;
    children?: (A11yTreeNodeMatch | string | RegExp)[];
};
type A11yTreeNodeStateForDiff = A11yTreeNodeState;
type A11yTreeNodeQueriesForDiff = {
    level?: number;
    value: {
        min?: number;
        max?: number;
        now?: number;
        text?: TextMatcher;
    };
};
type A11yTreeForDiff = {
    name?: TextMatcher;
    role?: HTMLElement['role'] | undefined;
    description?: TextMatcher;
    state?: A11yTreeNodeStateForDiff;
    queries?: A11yTreeNodeQueriesForDiff;
    children?: (A11yTreeForDiff | StaticText | Exclude<TextMatcher, 'string' | 'number'> | undefined)[];
};

type MatcherOptions = A11yTreeNodeContext;

declare global {
  namespace jest {
    interface Matchers<R> {
      /**
       * @description
       * Asserts that the accessibility tree for an element matches the expected structure.
       * An accessibility tree represents how a user agent (such as a screen reader) processes and communicates
       * accessibility information from the DOM. Use this matcher to check whether the pertinent accessibility
       * properties and relationships are correctly established, ensuring an accessible experience for users
       * of assistive technologies.
       *
       * The expected structure is defined by the `A11yTreeNodeMatch` interface, which can accommodate various
       * properties like role, name, description, state, and children.
       *
       * The second parameter allows for optional configuration settings
       *
       * @example
       * // Example with byRole hierarchy:
       * <nav aria-label="Main navigation">
       *   <ul>
       *     <li><a href="/home">Home</a></li>
       *     <li><a href="/about">About</a></li>
       *   </ul>
       *   <button disabled>Click me</button>
       * </nav>
       *
       * expect(screen.getByRole('navigation', { name: 'Main navigation' })).toHaveA11yTree(
       *   byRole('navigation', [
       *     byRole('list', [
       *       byRole('listitem', [
       *         byRole('link', { name: 'Home' }),
       *       ]),
       *       byRole('listitem', [
       *         byRole('link', { name: 'About' }),
       *       ]),
       *     ]),
       *     byRole('button', { name: 'Click me', disabled: true }),
       *   ])
       * );
       *
       * @see
       * - [W3C Accessibility Tree](https://www.w3.org/TR/wai-aria/#accessibility_tree)
       * - [ARIA in HTML](https://www.w3.org/TR/html-aria/)
       */
      toHaveA11yTree(match?: A11yTreeNodeMatch, options?: MatcherOptions): R;
    }
  }
}

export { A11yTreeNodeMatch as A, MatcherOptions as M, StaticText as S, TextMatcher as T, VirtualRoles as V, A11yTreeNodeStateMatch as a, A11yTreeNodeQueriesMatch as b, A11yTreeForDiff as c, A11yTreeNode as d, AsNonLandmarkRoles as e, ARIARoleDefinitionKeyExtended as f, A11yTreeNodeState as g };
